use std::net::SocketAddr;

use eyre::eyre;
use tracing::{debug, instrument};
use zerocopy::{AsBytes, LittleEndian};

use crate::packets::{Header, Packet, PacketKind};

type U16 = zerocopy::U16<LittleEndian>;
type U32 = zerocopy::U32<LittleEndian>;

#[derive(AsBytes)]
#[repr(C)]
struct DynIpUpdate {
    number: U32,
    pin: U16,
    port: U16,
}

/// # Errors
/// - the dyn ip server returns a malformed response or is unreachable
/// - the authentication fails
#[instrument]
pub async fn dyn_ip_update(
    server: &SocketAddr,
    number: u32,
    pin: u16,
    port: u16,
) -> eyre::Result<std::net::Ipv4Addr> {
    debug!(%number, %port, "starting dyn ip update");

    let mut packet = Packet {
        header: Header {
            kind: PacketKind::DynIpUpdate.raw(),
            length: 8,
        },
        data: Vec::new(),
    };

    packet.data.resize(packet.header.length as usize, 0);

    DynIpUpdate {
        number: number.into(),
        pin: pin.into(),
        port: port.into(),
    }
    .write_to(packet.data.as_mut_slice())
    .unwrap();

    let mut socket = tokio::net::TcpStream::connect(server).await?;

    let (mut reader, mut writer) = socket.split();

    packet.send(&mut writer).await?;
    packet.recv_into(&mut reader).await?;

    let result = match packet.kind() {
        PacketKind::DynIpUpdateResponse => Ok(<[u8; 4]>::try_from(packet.data)
            .map_err(|err| {
                eyre!(
                    "too little data for ip address. Need 4 bytes got {}",
                    err.len()
                )
            })?
            .into()),
        PacketKind::Error => {
            let first_zero = packet
                .data
                .iter()
                .enumerate()
                .find_map(|(i, x)| (*x == 0).then_some(i));

            return Err(eyre!(
                "{}",
                std::str::from_utf8(first_zero.map_or(&packet.data, |i| &packet.data[..i]),)?
            ));
        }

        _ => return Err(eyre!("server returned unexpected packet")),
    };

    debug!(?result, "finished dyn ip update");

    result
}
